from enum import Enum

import cv2
from sorcery import assigned_names

from draugr.opencv_utilities.namespaces.enums import (
    MorphShapeEnum,
    MorphTypeEnum,
    BorderTypeEnum,
)

__all__ = ["clean_up", "CleanUpMethod"]


class CleanUpMethod(Enum):
    close, open, nclose, nopen, erode, dilate, none = assigned_names()


def clean_up(img, method: CleanUpMethod = CleanUpMethod.open, **kwargs):
    filter_size = (
        max(int(img.shape[1] * 0.01), kwargs.get("min_filter_size", 3)),
        max(int(img.shape[0] * 0.01), kwargs.get("min_filter_size", 3)),
    )
    morph_shape = MorphShapeEnum.rect.value
    kernel = cv2.getStructuringElement(morph_shape, filter_size)

    if method == CleanUpMethod.close:
        img = cv2.morphologyEx(
            img,
            MorphTypeEnum.close.value,
            kernel,
            iterations=kwargs.get("iterations", 1),
            borderType=kwargs.get("borderType", None),  # BorderTypeEnum.constant.value,
            borderValue=kwargs.get("borderValue", 0),  # 255
        )
    elif method == CleanUpMethod.open:
        img = cv2.morphologyEx(
            img,
            MorphTypeEnum.open.value,
            kernel,
            iterations=kwargs.get("iterations", 1),
            borderType=kwargs.get("borderType", None),  # BorderTypeEnum.constant.value,
            borderValue=kwargs.get("borderValue", 0),  # 255
        )
    elif method == CleanUpMethod.nclose:
        img = cv2.morphologyEx(
            img,
            MorphTypeEnum.dilate.value,
            kernel,
            None,
            None,
            iterations=kwargs.get("iterations", 2),
            borderType=kwargs.get("borderType", BorderTypeEnum.reflect101.value),
        )
        img = cv2.morphologyEx(
            img,
            MorphTypeEnum.erode.value,
            kernel,
            None,
            None,
            iterations=kwargs.get("iterations", 2),
            borderType=kwargs.get("borderType", BorderTypeEnum.reflect101.value),
        )
    elif method == CleanUpMethod.erode:
        img = cv2.dilate(
            img,
            kernel,
            iterations=kwargs.get("iterations", 1),
        )
    elif method == CleanUpMethod.dilate:
        img = cv2.erode(
            img,
            kernel,
            iterations=kwargs.get("iterations", 1),
        )

    return img



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

__author__ = "Christian Heider Nielsen"
__doc__ = r"""

           Created on 17/07/2020
           """

from typing import Tuple

import PIL.Image
import numpy

from draugr.visualisation.pillow_utilities import np_array_to_pil_img

__all__ = ["resize_array"]

from warg import passes_kws_to


@passes_kws_to(PIL.Image.Image.resize)
def resize_array(x: numpy.ndarray, size: Tuple[int, int], **kwargs) -> numpy.ndarray:
    """ """
    assert x.ndim in [3, 4], "Only 3D and 4D Tensors allowed!"
    if isinstance(size, int):
        size = (size, size)
    if x.ndim == 4:  # 4D Tensor
        res = []
        for i in range(x.shape[0]):
            img = np_array_to_pil_img(x[i])
            img = img.resize(size, **kwargs)
            img = numpy.asarray(img, dtype="float32")
            img = numpy.expand_dims(img, axis=0)
            img /= 255.0
            res.append(img)
        res = numpy.concatenate(res)
        # res = numpy.expand_dims(res, axis=1) # TODO Wrong?
    else:  # 3D Tensor
        img = np_array_to_pil_img(x)
        img = img.resize(size, **kwargs)
        res = numpy.asarray(img, dtype="float32")
        res = numpy.expand_dims(res, axis=0)
        res /= 255.0
    return res


if __name__ == "__main__":
    print(resize_array(numpy.zeros((12, 10, 10, 3)), 5).shape)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

__author__ = "Christian Heider Nielsen"
__doc__ = ""
__all__ = ["launch_tensorboard"]

from pathlib import Path

from tensorboard import program  # IMPORT OUT HERE; POSSIBLE RAISE CONDITIONS

from draugr.python_utilities.sockets import is_port_in_use


def launch_tensorboard(log_dir: Path, port: int = 6006) -> str:
    """

    :param log_dir:
    :type log_dir:
    :param port:
    :type port:
    :return:
    :rtype:"""

    if is_port_in_use(port):
        raise RuntimeError(f"Port {port} is in use")

    tb = program.TensorBoard()
    # tb.configure(argv=['', '--logdir', log_dir, '--port', port])
    tb.configure(logdir=str(log_dir), port=port)
    tb._fix_mime_types()
    return tb.launch()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

__author__ = "Christian Heider Nielsen"
__doc__ = r"""

           Created on 17-09-2020
           """

from pathlib import Path

from .. import INCLUDE_PROJECT_READMES

if INCLUDE_PROJECT_READMES:
    with open(Path(__file__).parent / "README.md", "r") as this_init_file:
        __doc__ += this_init_file.read()

try:
    from .conversion import *
    from .singleton_engine import *
    from .matlab_conversion_utilities import *
except ImportError as ix:
    this_package_name = Path(__file__).parent.name
    this_package_reqs = (
        Path(__file__).parent.parent.parent
        / "requirements"
        / f"requirements_{this_package_name}.txt"
    )
    print(
        f"Make sure requirements is installed for {this_package_name}, see {this_package_reqs}"
    )  # TODO: PARSE WHAT is missing and print
    raise ix



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
from abc import ABC, abstractmethod

__author__ = "Christian Heider Nielsen"
__doc__ = r"""

           Created on 09/10/2019
           """
__all__ = ["BarWriterMixin"]


class BarWriterMixin(ABC):
    """
    Writer mixin that provides an interface for 'writing' bar charts"""

    @abstractmethod
    def bar(
        self,
        tag: str,
        values: list,
        step: int,
        y_error=None,
        x_labels=None,
        y_label="Probability",
        x_label="Action Categorical Distribution",
        **kwargs
    ) -> None:
        """

            :param values:
            :param y_error:
            :param x_labels:
            :param y_label:
            :param x_label:
        :param tag:
        :type tag:
        :param step:
        :type step:
        :param kwargs:
        :type kwargs:"""
        raise NotImplementedError



#!/usr/bin/env python3
# -*- coding: utf-8 -*-


__author__ = "heider"
__doc__ = r"""

           Created on 03/02/2022
           """

__all__ = []

import subprocess


def is_dark_mode_active():
    """
    Bad
    :return:
    :rtype:
    """
    try:
        out = subprocess.run(
            ["gsettings", "get", "org.gnome.desktop.interface", "gtk-theme"],
            capture_output=True,
        )
        # Gtk.Settings.get_default().set_property("gtk-theme-name", "MS-Windows-XP")
        stdout = out.stdout.decode()
    except Exception:
        return False

    if stdout.lower().strip()[1:-1].endswith("-dark"):
        return True
    return False


if __name__ == "__main__":
    print(is_dark_mode_active())



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
from typing import List, Tuple

import numpy
import torch

from torch.distributions import Categorical
from torch.nn import functional

from draugr.torch_utilities.tensors.to_tensor import to_tensor
from draugr.torch_utilities.architectures.mlp import MLP

__author__ = "Christian Heider Nielsen"
__doc__ = r"""
"""
__all__ = ["MultipleCategoricalMLP", "CategoricalMLP"]


class MultipleCategoricalMLP(MLP):
    @staticmethod
    def sample(distributions) -> Tuple:
        actions = [d.sample() for d in distributions][0]

        log_prob = [d.log_prob(action) for d, action in zip(distributions, actions)][0]

        actions = [a.to("cpu").numpy().tolist() for a in actions]
        return actions, log_prob

    @staticmethod
    def entropy(distributions) -> torch.tensor:
        return torch.mean(to_tensor([d.entropy() for d in distributions]))

    def forward(self, *x, **kwargs) -> List:
        out = super().forward(*x, **kwargs)
        outs = []
        for o in out:
            outs.append(Categorical(logits=functional.log_softmax(o, dim=-1)))

        return outs


class CategoricalMLP(MLP):
    def forward(self, *x, **kwargs) -> Categorical:
        return Categorical(
            logits=functional.log_softmax(super().forward(*x, **kwargs), dim=-1)
        )


if __name__ == "__main__":

    def multi_cat():
        s = (2, 2)
        a = (2, 2)
        model = MultipleCategoricalMLP(input_shape=s, output_shape=a)

        inp = to_tensor(numpy.random.rand(64, s[0]), device="cpu")
        print(model.sample(model(inp, inp)))

    def single_cat():
        s = (1, 2)
        a = (2,)
        model = CategoricalMLP(input_shape=s, output_shape=a)

        inp = to_tensor(numpy.random.rand(64, s[0]), device="cpu")
        inp2 = to_tensor(numpy.random.rand(64, s[1]), device="cpu")
        print(model(inp, inp2).sample())

    def single_cat2():
        s = (4,)
        a = (2,)
        model = CategoricalMLP(input_shape=s, output_shape=a)

        inp = to_tensor(numpy.random.rand(64, s[0]), device="cpu")
        print(model(inp).sample())

    multi_cat()
    single_cat()
    single_cat2()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

__author__ = "heider"
__doc__ = r"""

           Created on 01/02/2022
           """

import numpy

__all__ = ["mix_channels"]


def mix_channels(raster: numpy.ndarray) -> numpy.ndarray:
    # TODO: MAYBE ASSERT SHAPE?
    num_channels = raster.shape[-1]
    return numpy.dot(
        raster[..., :num_channels],
        numpy.ones(num_channels) / num_channels,
    )



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

__author__ = "Christian Heider Nielsen"
__doc__ = r"""

           Created on 25-05-2021
           """

from typing import Any

import numpy
import pandas

__all__ = ["add_index_level"]


def add_index_level(
    old_index: pandas.Index, value: Any, name: str = None, loc: int = 0
) -> pandas.MultiIndex:
    """
    Expand a (multi)index by adding a level to it.

    :param old_index: The index to expand
    :param name: The name of the new index level
    :param value: Scalar or list-like, the values of the new index level
    :param loc: Where to insert the level in the index, 0 is at the front, negative values count back from the rear end
    :return: A new multi-index with the new level added
    """

    def _handle_insert_loc(loc: int, n: int) -> int:
        """
        Computes the insert index from the right if loc is negative for a given size of n.
        """
        return n + loc + 1 if loc < 0 else loc

    loc = _handle_insert_loc(loc, len(old_index.names))
    old_index_df = old_index.to_frame()
    old_index_df.insert(loc, name, value)
    new_index_names = list(
        old_index.names
    )  # sometimes new index level names are invented when converting to a df,
    new_index_names.insert(loc, name)  # here the original names are reconstructed
    new_index = pandas.MultiIndex.from_frame(old_index_df, names=new_index_names)
    return new_index


if __name__ == "__main__":

    def stest_add_index_level() -> None:
        """
        :rtype: None
        """
        df = pandas.DataFrame(data=numpy.random.normal(size=(6, 3)))
        i1 = add_index_level(df.index, "foo")

        # it does not invent new index names where there are missing
        assert [None, None] == i1.names

        # the new level values are added
        assert numpy.all(i1.get_level_values(0) == "foo")
        assert numpy.all(i1.get_level_values(1) == df.index)

        # it does not invent new index names where there are missing
        i2 = add_index_level(i1, ["x", "y"] * 3, name="xy", loc=2)
        i3 = add_index_level(i2, ["a", "b", "c"] * 2, name="abc", loc=-1)
        assert [None, None, "xy", "abc"] == i3.names

        # the new level values are added
        assert numpy.all(i3.get_level_values(0) == "foo")
        assert numpy.all(i3.get_level_values(1) == df.index)
        assert numpy.all(i3.get_level_values(2) == ["x", "y"] * 3)
        assert numpy.all(i3.get_level_values(3) == ["a", "b", "c"] * 2)

        # df.index = i3
        # print()
        # print(df)

    stest_add_index_level()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
__author__ = "Christian Heider Nielsen"

import numpy
import torch

__all__ = ["orthogonal_weights"]


def orthogonal_weights(shape, scale: float = 1.0) -> torch.tensor:
    r"""PyTorch port of ortho_init from baselines.a2c.utils"""
    shape = tuple(shape)

    if len(shape) == 2:
        flat_shape = shape[1], shape[0]
    elif len(shape) == 4:
        flat_shape = (numpy.prod(shape[1:]), shape[0])
    else:
        raise NotImplementedError

    a = numpy.random.normal(0.0, 1.0, flat_shape)
    u, _, v = numpy.linalg.svd(a, full_matrices=False)
    q = u if u.shape == flat_shape else v
    q = q.transpose().copy().reshape(shape)

    if len(shape) == 2:
        return torch.from_numpy((scale * q).astype(numpy.float32))
    if len(shape) == 4:
        return torch.from_numpy(
            (scale * q[:, : shape[1], : shape[2]]).astype(numpy.float32)
        )



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

__author__ = "Christian Heider Nielsen"
__doc__ = r"""



           Created on 17-03-2021
           """

from draugr.tensorboard_utilities.exporting.event_export import TensorboardEventExporter


class TensorboardEventExporterDatabase(TensorboardEventExporter):
    pass



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

__author__ = "Christian Heider Nielsen"
__doc__ = r"""

           Created on 01/03/2020
           """

from abc import abstractmethod
from typing import Dict, Tuple

from torch.utils.data import Dataset

__all__ = ["SupervisedDataset"]

from draugr.numpy_utilities.datasets.splitting import (
    SplitEnum,
    SplitIndexer,
)
from warg import drop_unused_kws


class SupervisedDataset(Dataset):
    """
    Supervised Dataset is comprised of separate Splits"""

    @drop_unused_kws
    def __init__(self):
        pass

    @property
    def split_names(self) -> Dict[SplitEnum, str]:
        """

        :return:
        :rtype:"""
        return SplitIndexer.default_split_names

    @property
    @abstractmethod
    def response_shape(self) -> Tuple[int, ...]:
        """ """
        raise NotImplementedError

    @property
    @abstractmethod
    def predictor_shape(self) -> Tuple[int, ...]:
        """ """
        raise NotImplementedError

    @abstractmethod
    def __len__(self):
        raise NotImplementedError

    @abstractmethod
    def __getitem__(self, index: int):
        raise NotImplementedError


if __name__ == "__main__":
    print(SplitIndexer(521))
    print(SplitIndexer(2512).unnormalised(123))



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
from draugr.python_utilities.styling import (
    COLORS,
    PrintStyle,
    generate_style,
    get_terminal_size,
    scale,
)

__author__ = "Christian Heider Nielsen"

from typing import Sequence, Dict

import numpy

# sys.stdout.write(generate_style(u'Draugr Ûnicöde Probe\n', underline=True, italic=True))

__all__ = ["terminal_plot", "styled_terminal_plot_stats_shared_x"]


def terminal_plot(
    y: Sequence,
    *,
    x: Sequence = None,
    title: str = "Values",
    rows=None,
    columns=None,
    percent_size=(0.80, 0.80),
    x_offsets=(1, 1),
    y_offsets=(1, 1),
    printer=print,
    print_summary=True,
    plot_character="\u2981",
    print_style: PrintStyle = None,
    border_size=1,
):
    """
    x, y list of values on x- and y-axis
    plot those values within canvas size (rows and columns)"""

    num_y = len(y)
    if num_y == 0:
        return

    if x:
        if len(x) != num_y:
            raise ValueError(
                f"x argument must match the length of y, got x:{len(x)} and y:{num_y}"
            )
    else:
        x = range(num_y)

    if not rows or not columns:
        terminal_size = get_terminal_size()
        columns = terminal_size.columns
        rows = terminal_size.rows

    if percent_size:
        columns, rows = int(columns * percent_size[0]), int(rows * percent_size[1])

    # Scale points such that they fit on canvas
    drawable_columns = columns - sum(x_offsets) - border_size * 2
    drawable_rows = rows - sum(y_offsets) - border_size * 2

    x_scaled = scale(x, drawable_columns)
    y_scaled = scale(y, drawable_rows)

    # Create empty canvas
    canvas = [[" " for _ in range(columns)] for _ in range(rows)]

    # Create borders
    for iy in range(1, rows - 1):
        canvas[iy][0] = "\u2502"
        canvas[iy][columns - 1] = "\u2502"
    for ix in range(1, columns - 1):
        canvas[0][ix] = "\u2500"
        canvas[rows - 1][ix] = "\u2500"

    canvas[0][0] = "\u250c"
    canvas[0][columns - 1] = "\u2510"
    canvas[rows - 1][0] = "\u2514"
    canvas[rows - 1][columns - 1] = "\u2518"

    # Add scaled points to canvas
    y_offsets_start = border_size + y_offsets[0]
    x_offsets_start = border_size + x_offsets[0]
    for ix, iy in zip(x_scaled, y_scaled):
        y_ = y_offsets_start + (drawable_rows - iy)
        x_ = x_offsets_start + ix
        canvas[y_][x_] = plot_character

    print("\n")
    # Print rows of canvas
    for row in ["".join(row) for row in canvas]:
        if print_style:
            printer(print_style(row))
        else:
            printer(row)

    # Print scale
    if print_summary:
        summary = (
            f"{title} - (min, max): x({min(x)}, {max(x)}), y({min(y)}, {max(y)})\n"
        )
        if print_style:
            printer(print_style(summary))
        else:
            printer(summary)


def styled_terminal_plot_stats_shared_x(stats, *, styles=None, **kwargs):
    """

    :param stats:
    :type stats:
    :param styles:
    :type styles:
    :param kwargs:
    :type kwargs:
    :return:
    :rtype:"""
    if styles is None:
        styles = [
            generate_style(color=color, highlight=True)
            for color, _ in zip(COLORS.keys(), range(len(stats)))
        ]
    return terminal_plot_stats_shared_x(stats, styles=styles, **kwargs)


def terminal_plot_stats_shared_x(
    stats: Dict,
    *,
    x: Sequence = None,
    styles=None,
    printer=print,
    margin=0.25,
    summary=True,
):
    """

    :param stats:
    :type stats:
    :param x:
    :type x:
    :param styles:
    :type styles:
    :param printer:
    :type printer:
    :param margin:
    :type margin:
    :param summary:
    :type summary:"""
    num_stats = len(stats)

    y_size = (1 - margin) / num_stats

    if styles:
        if len(styles) != num_stats:
            raise ValueError(
                f"styles argument must match the length of stats, got styles:{len(styles)} and "
                f"stats:{num_stats}"
            )
    else:
        styles = [None for _ in range(num_stats)]

    for (key, stat), sty in zip(stats.items(), styles):
        terminal_plot(
            stat.running_value,
            title=key,
            x=x,
            printer=printer,
            print_style=sty,
            percent_size=(1, y_size),
            print_summary=summary,
        )


if __name__ == "__main__":
    terminal_plot(numpy.tile(range(9), 4), plot_character="o")



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

__author__ = "Christian Heider Nielsen"
__doc__ = r"""

           Created on 17/07/2020
           """

__all__ = ["recursive_flatten_numpy", "MyNdArray"]

from typing import Sequence

import numpy


def recursive_flatten_numpy(nested: Sequence) -> numpy.ndarray:
    """ """
    return numpy.array(nested).ravel()


class MyNdArray(numpy.ndarray):
    """
    Small extension
    """

    @property
    def conjugate_transpose(self):
        """

        :return:
        """
        return self.conj().T


if __name__ == "__main__":
    print(recursive_flatten_numpy([[numpy.zeros((2, 2))], [numpy.zeros((2, 2))]]))
    print(recursive_flatten_numpy(numpy.zeros((2, 2))))
    print(numpy.random.random((3, 3)).view(MyNdArray).conjugate_transpose)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
from abc import ABC, abstractmethod
from enum import Enum
from typing import Union

import numpy
import torch
from PIL import Image
from sorcery import assigned_names

__author__ = "Christian Heider Nielsen"
__doc__ = r"""

           Created on 09/10/2019
           """
__all__ = ["ImageWriterMixin"]


class ImageWriterMixin(ABC):
    """
    Writer subclass that provides an interface for 'writing' images
    +Bonus points for handling other multichannel rasters C!=1 and C!=3
    """

    class MultiChannelMethodEnum(
        Enum
    ):  # TODO: MAYBE RENAME TO something with REDUCTIONMETHODS
        seperate, mix, project = assigned_names()

    @abstractmethod
    def image(
        self,
        tag: str,
        data: Union[numpy.ndarray, torch.Tensor, Image.Image],
        step,
        *,
        dataformats: str = "NCHW",
        **kwargs
    ) -> None:
        """

        :param tag:
        :type tag:
        :param data:
        :type data:
        :param step:
        :type step:
        :param dataformats:
        :type dataformats:
        :param kwargs:
        :type kwargs:"""
        raise NotImplementedError



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

__author__ = "Christian Heider Nielsen"
__doc__ = r"""

           Created on 07/07/2020
           """

import functools
from collections import OrderedDict
from typing import Tuple

import torch
from torch import nn

__all__ = ["IntermediateLayerGetter"]


class IntermediateLayerGetter:
    """ """

    def __init__(
        self,
        model: torch.nn.Module,
        return_layers: dict = None,
    ):
        """
        Wraps a Pytorch module to get intermediate values, eg for getting intermediate activations

        Arguments:
        model {nn.module} -- The Pytorch module to call
        return_layers {dict} -- Dictionary with the selected submodules
        to return the output (format: {[current_module_name]: [desired_output_name]},
        current_module_name can be a nested submodule, e.g. submodule1.submodule2.submodule3)

        Returns:
        (mid_outputs {OrderedDict}, model_output {any}) -- mid_outputs keys are
        your desired_output_name (s) and their values are the returned tensors
        of those submodules (OrderedDict([(desired_output_name,tensor(...)), ...).

        In case a submodule is called more than one time, all it's outputs are
        stored in a list."""
        self._model = model
        if return_layers:
            self.return_layers = return_layers.items()
        else:
            self.return_layers = {k: k for k, v in model.named_modules()}.items()

    @staticmethod
    def reduce_getattr(obj, attr, *args):
        """
        # using wonder's beautiful simplification: https://stackoverflow.com/questions/31174295/getattr-and-setattr-on-nested-objects/31174427?noredirect=1#comment86638618_31174427

        :param obj:
        :type obj:
        :param attr:
        :type attr:
        :param args:
        :type args:
        :return:
        :rtype:"""

        def _getattr(obj, attr):
            return getattr(obj, attr, *args)

        return functools.reduce(_getattr, (obj, *attr.split(".")))

    def __call__(self, *args, **kwargs) -> Tuple:
        ret = OrderedDict()
        handles = []
        for name, new_name in self.return_layers:
            if name == "":
                continue  # TODO: Fail maybe?
            layer = IntermediateLayerGetter.reduce_getattr(self._model, name)

            if isinstance(layer, torch.nn.Module):  # Should be a torch module!

                def hook(
                    module,
                    input,
                    output,
                    *,
                    new_name_=new_name,  # Hack for new func, otherwise func is overriden. # BUG?
                ):
                    """

                              :param new_name_:
                    :param module:
                    :type module:
                    :param input:
                    :type input:
                    :param output:
                    :type output:"""
                    if new_name_ in ret:
                        cur_val = ret[new_name_]
                        if type(cur_val) is list:
                            ret[new_name_].append(output)
                        else:
                            ret[new_name_] = [cur_val, output]
                    else:
                        ret[new_name_] = output

                try:
                    h = layer.register_forward_hook(hook)
                    handles.append(h)
                except AttributeError as e:
                    raise AttributeError(f"Module {name} not found")
            else:
                raise AttributeError(
                    f"Requested module activation with {name} was not a module but {type(layer)}"
                )

        output = self._model(*args, **kwargs)

        for h in handles:
            h.remove()

        return ret, output


if __name__ == "__main__":

    def adsad() -> None:
        """
        :rtype: None
        """

        class Model(nn.Module):
            """ """

            def __init__(self):
                super().__init__()

                self.fc1 = nn.Linear(2, 2)
                self.fc2 = nn.Linear(2, 2)
                self.nested = nn.Sequential(
                    nn.Sequential(nn.Linear(2, 2), nn.Linear(2, 3)), nn.Linear(3, 1)
                )
                self.interaction_idty = (
                    nn.Identity()
                )  # Simple trick for operations not performed as modules

            def forward(self, x):
                """

                :param x:
                :type x:
                :return:
                :rtype:"""
                x1 = self.fc1(x)
                x2 = self.fc2(x)

                interaction = x1 * x2
                self.interaction_idty(interaction)
                return self.nested(interaction)

        model = Model()
        return_layers = {
            "fc2": "fc2",
            "nested.0.1": "nested",
            "interaction_idty": "interaction",
        }
        mid_getter = IntermediateLayerGetter(model, return_layers=return_layers)
        mid_outputs, model_output = mid_getter(torch.randn(1, 2))

        print(model_output)
        print(mid_outputs)

    def adsad2() -> None:
        """
        :rtype: None
        """

        class Model(nn.Module):
            """ """

            def __init__(self):
                super().__init__()

                self.fc1 = nn.Linear(2, 2)
                self.fc2 = nn.Linear(2, 2)
                self.nested = nn.Sequential(
                    nn.Sequential(nn.Linear(2, 2), nn.Linear(2, 3)), nn.Linear(3, 1)
                )
                self.interaction_idty = (
                    nn.Identity()
                )  # Simple trick for operations not performed as modules

            def forward(self, x):
                """

                :param x:
                :type x:
                :return:
                :rtype:"""
                x1 = self.fc1(x)
                x2 = self.fc2(x)

                interaction = x1 * x2
                self.interaction_idty(interaction)
                return self.nested(interaction)

        model = Model()

        mid_getter = IntermediateLayerGetter(model)
        mid_outputs, model_output = mid_getter(torch.randn(1, 2))

        print(model_output)
        print(mid_outputs)

    adsad()
    # adsad2()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

__author__ = "Christian Heider Nielsen"
__doc__ = r"""

           Created on 14-02-2021
           """

from typing import Sequence

import numpy

__all__ = ["get_strides_f", "get_strides_c"]


def get_strides_f(shape: Sequence) -> object:
    """Get strides of a F like array, for numpy array need to multiply by itemsize

    Parameters
    ----------
    shape : tuple of int or iterable
      shape of the array
    Returns
    -------
    s :  tuple of int or iterable
      strides of the array

    Examples
    --------
    >>> get_strides_f((2, 3, 3))
    [1, 2, 6]

    References
    ----------
    https://docs.scipy.org/doc/numpy/reference/arrays.ndarray.html"""

    # $s_k = \Prod_{j=0}^{k-1} d_j$
    s = numpy.cumprod((1,) + shape[0:-1]).tolist()

    return s


def get_strides_c(shape):
    """Get strides of a C like array. For numpy array need to be multiply by itemsize

    Parameters
    ----------
    shape : tuple of int or iterable
      shape of the array
    Returns
    -------
    s :  tuple of int or iterable
      strides of the array

    Examples
    --------
    >>> get_strides_c((2, 3, 3))
    [9, 3, 1]
    References
    ----------
    https://docs.scipy.org/doc/numpy/reference/arrays.ndarray.html"""

    n = len(shape)

    # $s_k = \Prod_{j=k+1}^{N-1} d_j$ with shape[j] <=> $d_j$
    s = [0] * n
    for k in reversed(range(0, n)):
        if k == n - 1:
            s[k] = 1
        else:
            s[k] = shape[k + 1] * s[k + 1]

    return s



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

__author__ = "Christian Heider Nielsen"
__doc__ = r"""

           Created on 20-01-2021
           """

import numpy
from matplotlib import cm, pyplot

__all__ = ["plot_gradients_3d"]


def plot_gradients_3d() -> pyplot.Figure:
    """ """
    fig = pyplot.figure()
    ax = fig.gca(projection="3d")
    x = numpy.arange(-5, 5, 0.25)
    y = numpy.arange(-5, 5, 0.25)
    x, y = numpy.meshgrid(x, y)
    r = numpy.sqrt(x**2 + y**2)
    z = numpy.sin(r)
    gx, gy = numpy.gradient(z)  # gradients with respect to x and y
    g = (gx**2 + gy**2) ** 0.5  # gradient magnitude
    n = g / g.max()  # normalize 0..1
    ax.plot_surface(
        x,
        y,
        z,
        rstride=1,
        cstride=1,
        facecolors=cm.jet(n),
        linewidth=0,
        antialiased=False,
        shade=False,
    )
    return fig


if __name__ == "__main__":
    plot_gradients_3d()
    pyplot.show()



from enum import Enum


class TaskCreationEnum(Enum):
    """ """

    TASK_VALIDATE_ONLY = 1
    TASK_CREATE = 2
    TASK_UPDATE = 4
    TASK_CREATE_OR_UPDATE = 6  # If task already exists, it will be updated
    TASK_DISABLE = 8
    TASK_DONT_ADD_PRINCIPAL_ACE = 10
    TASK_IGNORE_REGISTRATION_TRIGGERS = 20



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

__author__ = "heider"
__doc__ = r"""

           Created on 01/02/2022
           """

__all__ = [
    "MarkerTypeEnum",
    "DistanceTypeEnum",
    "HoughModeEnum",
    "WindowFlagEnum",
    "ContourRetrievalModeEnum",
    "ContourApproximationModeEnum",
    "MouseEventEnum",
    "FontEnum",
    "DataTypeEnum",
    "ComparisonEnum",
    "LineTypeEnum",
    "MorphTypeEnum",
    "MorphShapeEnum",
    "DistanceTransformMaskEnum",
    "DistanceTransformLabelTypeEnum",
    "KmeansEnum",
    "CameraPropertyEnum",
    "RectanglesIntersectTypes",
    "BorderTypeEnum",
    "VideoCaptureAPIEnum",
]

from enum import Enum

import cv2


class ComparisonEnum(Enum):
    eq = cv2.CMP_EQ  # equal
    gt = cv2.CMP_GT  # greater than
    ge = cv2.CMP_GE  # # greater equal
    lt = cv2.CMP_LT  # less than
    le = cv2.CMP_LE  # less equal
    ne = cv2.CMP_NE  # unequal


class DataTypeEnum(Enum):
    # Unsigned 8bits # 	uchar	0 ~ 255
    u8 = cv2.CV_8U
    u8c1 = cv2.CV_8UC1
    u8c2 = cv2.CV_8UC2
    u8c3 = cv2.CV_8UC3
    u8c4 = cv2.CV_8UC4
    # u8cN
    # Signed 8bits	char	-128 ~ 127
    s8 = cv2.CV_8S
    s8c1 = cv2.CV_8SC1
    s8c2 = cv2.CV_8SC2
    s8c3 = cv2.CV_8SC3
    s8c4 = cv2.CV_8SC4
    # s8cN
    # Unsigned 16bits	ushort	0 ~ 65535
    u16 = cv2.CV_16U
    u16c1 = cv2.CV_16UC1
    u16c2 = cv2.CV_16UC2
    u16c3 = cv2.CV_16UC3
    u16c4 = cv2.CV_16UC4
    # u16cN
    # Signed 16bits	short	-32768 ~ 32767
    s16 = cv2.CV_16S
    s16c1 = cv2.CV_16SC1
    s16c2 = cv2.CV_16SC2
    s16c3 = cv2.CV_16SC3
    s16c4 = cv2.CV_16SC4
    # s16cN
    # Signed 32bits	int	-2147483648 ~ 2147483647
    s32 = cv2.CV_32S
    s32c1 = cv2.CV_32SC1
    s32c2 = cv2.CV_32SC2
    s32c3 = cv2.CV_32SC3
    s32c4 = cv2.CV_32SC4
    # s32cN
    # Float 32bits	float	-1.18e-38 ~ 3.40e-38
    f32 = cv2.CV_32F
    f32c1 = cv2.CV_32FC1
    f32c2 = cv2.CV_32FC2
    f32c3 = cv2.CV_32FC3
    f32c4 = cv2.CV_32FC4
    # f32cN
    # Double 64bits	double	-1.7e+308 ~ +1.7e+308
    f64 = cv2.CV_64F
    f64c1 = cv2.CV_64FC1
    f64c2 = cv2.CV_64FC2
    f64c3 = cv2.CV_64FC3
    f64c4 = cv2.CV_64FC4
    # f64cN


class VideoCaptureAPIEnum(Enum):
    """



    VideoCapture API backends identifier.
    """

    any = cv2.CAP_ANY
    # Auto detect == 0.

    vfw = cv2.CAP_VFW
    # Video For Windows (platform native)

    v4l = cv2.CAP_V4L
    # V4L/V4L2 capturing support via libv4l.

    v4l2 = cv2.CAP_V4L2
    # Same as CAP_V4L.

    firewire = cv2.CAP_FIREWIRE
    # IEEE 1394 drivers.

    fireware = cv2.CAP_FIREWARE
    # Same as CAP_FIREWIRE.

    ieee1394 = cv2.CAP_IEEE1394
    # Same as CAP_FIREWIRE.

    dc1394 = cv2.CAP_DC1394
    # Same as CAP_FIREWIRE.

    cmu1394 = cv2.CAP_CMU1394
    # Same as CAP_FIREWIRE.

    qt = cv2.CAP_QT
    # QuickTime.

    unicap = cv2.CAP_UNICAP
    # Unicap drivers.

    dshow = cv2.CAP_DSHOW
    # DirectShow (via videoInput)

    pvapi = cv2.CAP_PVAPI
    # PvAPI, Prosilica GigE SDK.

    openni = cv2.CAP_OPENNI
    # OpenNI (for Kinect)

    openni_asus = cv2.CAP_OPENNI_ASUS
    # OpenNI (for Asus Xtion)

    android = cv2.CAP_ANDROID
    # Android - not used.

    xiapi = cv2.CAP_XIAPI
    # XIMEA Camera API.

    avfoundation = cv2.CAP_AVFOUNDATION
    # AVFoundation framework for iOS (OS X Lion will have the same API)

    giganetix = cv2.CAP_GIGANETIX
    # Smartek Giganetix GigEVisionSDK.

    msmf = cv2.CAP_MSMF
    # Microsoft Media Foundation (via videoInput)

    winrt = cv2.CAP_WINRT
    # Microsoft Windows Runtime using Media Foundation.

    intelperc = cv2.CAP_INTELPERC
    # Intel Perceptual Computing SDK.

    openni2 = cv2.CAP_OPENNI2
    # OpenNI2 (for Kinect)

    openni2_asus = cv2.CAP_OPENNI2_ASUS
    # OpenNI2 (for Asus Xtion and Occipital Structure sensors)

    gphoto2 = cv2.CAP_GPHOTO2
    # gPhoto2 connection

    gstreamer = cv2.CAP_GSTREAMER
    # GStreamer.

    ffmpeg = cv2.CAP_FFMPEG
    # Open and record video file or stream using the FFMPEG library.

    images = cv2.CAP_IMAGES
    # OpenCV Image Sequence (e.g. img_%02d.jpg)

    aravis = cv2.CAP_ARAVIS
    # Aravis SDK.

    opencv_mjpeg = cv2.CAP_OPENCV_MJPEG
    # Built-in OpenCV MotionJPEG codec.

    intel_mfx = cv2.CAP_INTEL_MFX
    # Intel MediaSDK.

    xine = cv2.CAP_XINE
    # XINE engine (Linux)


class CameraPropertyEnum(Enum):
    settings = cv2.CAP_PROP_SETTINGS
    auto_focus = cv2.CAP_PROP_AUTOFOCUS
    pos_msec = (
        cv2.CAP_PROP_POS_MSEC
    )  # Current position of the video file in milliseconds.
    pos_frames = (
        cv2.CAP_PROP_POS_FRAMES
    )  # 0-based index of the frame to be decoded/captured next.
    avi_ratio = (
        cv2.CAP_PROP_POS_AVI_RATIO
    )  # Relative position of the video file: 0 - start of the film, 1 - end of the film.
    frame_width = cv2.CAP_PROP_FRAME_WIDTH  # Width of the frames in the video stream.
    frame_height = (
        cv2.CAP_PROP_FRAME_HEIGHT
    )  # Height of the frames in the video stream.
    fps = cv2.CAP_PROP_FPS  # Frame rate.
    fourcc = cv2.CAP_PROP_FOURCC  # 4-character code of codec.
    frame_count = cv2.CAP_PROP_FRAME_COUNT  # Number of frames in the video file.
    format = cv2.CAP_PROP_FORMAT  # Format of the Mat objects returned by retrieve() .
    mode = (
        cv2.CAP_PROP_MODE
    )  # Backend-specific value indicating the current capture mode.
    brightness = cv2.CAP_PROP_BRIGHTNESS  # Brightness of the image (only for cameras).
    contrast = cv2.CAP_PROP_CONTRAST  # Contrast of the image (only for cameras).
    saturation = cv2.CAP_PROP_SATURATION  # Saturation of the image (only for cameras).
    hue = cv2.CAP_PROP_HUE  # Hue of the image (only for cameras).
    gain = cv2.CAP_PROP_GAIN  # Gain of the image (only for cameras).
    exposure = cv2.CAP_PROP_EXPOSURE  # Exposure (only for cameras).
    convert_rgb = (
        cv2.CAP_PROP_CONVERT_RGB
    )  # Boolean flags indicating whether images should be converted to RGB.
    white_balance_blue = cv2.CAP_PROP_WHITE_BALANCE_BLUE_U  # Currently unsupported
    white_balance_red = cv2.CAP_PROP_WHITE_BALANCE_RED_V  # Currently unsupported
    white_balance_ios = cv2.CAP_PROP_IOS_DEVICE_WHITEBALANCE  # Currently unsupported
    rectification = (
        cv2.CAP_PROP_RECTIFICATION
    )  # Rectification flag for stereo cameras (note: only supported by DC1394 v 2.x backend currently)
    monochrome = cv2.CAP_PROP_MONOCHROME
    sharpness = cv2.CAP_PROP_SHARPNESS
    gamma = cv2.CAP_PROP_GAMMA
    temperature = cv2.CAP_PROP_TEMPERATURE
    trigger = cv2.CAP_PROP_TRIGGER
    trigger_delay = cv2.CAP_PROP_TRIGGER_DELAY
    zoom = cv2.CAP_PROP_ZOOM
    focus = cv2.CAP_PROP_FOCUS
    guid = cv2.CAP_PROP_GUID
    iso_speed = cv2.CAP_PROP_ISO_SPEED
    backlight = cv2.CAP_PROP_BACKLIGHT
    pan = cv2.CAP_PROP_PAN
    tilt = cv2.CAP_PROP_TILT
    roll = cv2.CAP_PROP_ROLL
    iris = cv2.CAP_PROP_IRIS
    buffer_size = cv2.CAP_PROP_BUFFERSIZE
    sar_num = cv2.CAP_PROP_SAR_NUM
    sar_den = cv2.CAP_PROP_SAR_DEN
    backend = cv2.CAP_PROP_BACKEND
    channel = cv2.CAP_PROP_CHANNEL
    auto_wb = cv2.CAP_PROP_AUTO_WB
    wb_temperature = cv2.CAP_PROP_WB_TEMPERATURE
    codec_pixel_format = cv2.CAP_PROP_CODEC_PIXEL_FORMAT
    bitrate = cv2.CAP_PROP_BITRATE
    orientation_meta = cv2.CAP_PROP_ORIENTATION_META
    orientation_auto = cv2.CAP_PROP_ORIENTATION_AUTO
    open_timeout_msec = cv2.CAP_PROP_OPEN_TIMEOUT_MSEC
    read_timeout_msec = cv2.CAP_PROP_READ_TIMEOUT_MSEC
    stream_open_time_usec = cv2.CAP_PROP_STREAM_OPEN_TIME_USEC
    video_total_channels = cv2.CAP_PROP_VIDEO_TOTAL_CHANNELS
    video_stream = cv2.CAP_PROP_VIDEO_STREAM
    audio_stream = cv2.CAP_PROP_AUDIO_STREAM
    audio_pos = cv2.CAP_PROP_AUDIO_POS
    audio_shift_nsec = cv2.CAP_PROP_AUDIO_SHIFT_NSEC
    audio_data_depth = cv2.CAP_PROP_AUDIO_DATA_DEPTH
    audio_samples_per_second = cv2.CAP_PROP_AUDIO_SAMPLES_PER_SECOND
    audio_base_index = cv2.CAP_PROP_AUDIO_BASE_INDEX
    audio_total_channels = cv2.CAP_PROP_AUDIO_TOTAL_CHANNELS
    audio_total_streams = cv2.CAP_PROP_AUDIO_TOTAL_STREAMS
    audio_synchronize = cv2.CAP_PROP_AUDIO_SYNCHRONIZE
    lrf_has_key_frame = cv2.CAP_PROP_LRF_HAS_KEY_FRAME
    codec_extradata_index = cv2.CAP_PROP_CODEC_EXTRADATA_INDEX


class MarkerTypeEnum(Enum):
    cross = cv2.MARKER_CROSS  # A crosshair marker shape.
    tilted_cross = cv2.MARKER_TILTED_CROSS  # A 45 degree tilted crosshair marker shape.
    star = (
        cv2.MARKER_STAR
    )  # A star marker shape, combination of cross and tilted cross.
    diamond = cv2.MARKER_DIAMOND  # A diamond marker shape.
    square = cv2.MARKER_SQUARE  # A square marker shape.
    triangle_up = cv2.MARKER_TRIANGLE_UP  # An upwards pointing triangle marker shape.
    triangle_down = (
        cv2.MARKER_TRIANGLE_DOWN
    )  # A downwards pointing triangle marker shape.


class DistanceTypeEnum(Enum):
    user = cv2.DIST_USER  # User defined distance.
    l1 = cv2.DIST_L1  # distance = |x1-x2| + |y1-y2|
    l2 = cv2.DIST_L2  # the simple euclidean distance
    c = cv2.DIST_C  # distance = max(|x1-x2|,|y1-y2|)
    l12 = cv2.DIST_L12  # L1-L2 metric: distance = 2(sqrt(1+x*x/2) - 1))
    fair = cv2.DIST_FAIR  # distance = c^2(|x|/c-log(1+|x|/c)), c = 1.3998
    welsch = cv2.DIST_WELSCH  # distance = c^2/2(1-exp(-(x/c)^2)), c = 2.9846
    huber = cv2.DIST_HUBER  # distance = |x|<c ? x^2/2 : c(|x|-c/2), c=1.345


class DistanceTransformLabelTypeEnum(Enum):
    """
    distanceTransform algorithm flags
    """

    ccomp = cv2.DIST_LABEL_CCOMP
    """each connected component of zeros in src (as well as all the non-zero pixels closest to the connected component) will be assigned the same label
"""

    pixel = cv2.DIST_LABEL_PIXEL
    """each zero pixel (and all the non-zero pixels closest to it) gets its own label."""


class BorderTypeEnum(Enum):
    """
    Various border types, image boundaries are denoted with |
    """

    constant = cv2.BORDER_CONSTANT  # iiiiii|abcdefgh|iiiiiii with some specified i
    replicate = cv2.BORDER_REPLICATE  # aaaaaa|abcdefgh|hhhhhhh
    reflect = cv2.BORDER_REFLECT  # fedcba|abcdefgh|hgfedcb
    wrap = cv2.BORDER_WRAP  # cdefgh|abcdefgh|abcdefg
    reflect_101 = cv2.BORDER_REFLECT_101  # gfedcb|abcdefgh|gfedcba
    transparent = cv2.BORDER_TRANSPARENT  # uvwxyz|abcdefgh|ijklmno
    reflect101 = cv2.BORDER_REFLECT101  # same as BORDER_REFLECT_101
    default = cv2.BORDER_DEFAULT  # same as BORDER_REFLECT_101
    isolated = cv2.BORDER_ISOLATED  # do not look outside of ROI


class DistanceTransformMaskEnum(Enum):
    """
    Mask size for distance transform.
    """

    mask_3 = cv2.DIST_MASK_3
    mask_5 = cv2.DIST_MASK_5
    mask_precise = cv2.DIST_MASK_PRECISE


class KmeansEnum(Enum):
    random_centers = (
        cv2.KMEANS_RANDOM_CENTERS
    )  # Select random initial centers in each attempt.
    pp_centers = (
        cv2.KMEANS_PP_CENTERS
    )  # Use kmeans++ center initialization by Arthur and Vassilvitskii [Arthur2007].
    use_initial_labels = (
        cv2.KMEANS_USE_INITIAL_LABELS
    )  # During the first (and possibly the only) attempt, use the user-supplied labels instead of computing them from the initial centers. For the second and further attempts, use the random or semi-random centers. Use one of KMEANS_*_CENTERS flag to specify the exact method.


class HoughModeEnum(Enum):
    standard = (
        cv2.HOUGH_STANDARD
    )  # classical or standard Hough transform. Every line is represented by two floating-point numbers (ρ,θ) , where ρ is a distance between (0,0) point and the line, and θ is the angle between x-axis and the normal to the line. Thus, the matrix must be (the created sequence will be) of CV_32FC2 type
    gradient = cv2.HOUGH_GRADIENT  # basically 21HT
    probabilistic = (
        cv2.HOUGH_PROBABILISTIC
    )  # probabilistic Hough transform (more efficient in case if the picture contains a few long linear segments). It returns line segments rather than the whole line. Each segment is represented by starting and ending points, and the matrix must be (the created sequence will be) of the CV_32SC4 type.
    multi_scale = (
        cv2.HOUGH_MULTI_SCALE
    )  # multi-scale variant of the classical Hough transform. The lines are encoded the same way as HOUGH_STANDARD.
    gradient_alternative = (
        cv2.HOUGH_GRADIENT_ALT
    )  # variation of HOUGH_GRADIENT to get better accuracy


class WindowFlagEnum(Enum):
    normal = (
        cv2.WINDOW_NORMAL
    )  # the user can resize the window (no constraint) / also use to switch a fullscreen window to a normal size.
    autosize = (
        cv2.WINDOW_AUTOSIZE
    )  # the user cannot resize the window, the size is constrainted by the image displayed.
    opengl = cv2.WINDOW_OPENGL  # window with opengl support.
    fullscreen = cv2.WINDOW_FULLSCREEN  # change the window to fullscreen.
    free_ratio = (
        cv2.WINDOW_FREERATIO
    )  # the image expends as much as it can (no ratio constraint).
    keep_ratio = cv2.WINDOW_KEEPRATIO  # the ratio of the image is respected.
    gui_expanded = cv2.WINDOW_GUI_EXPANDED  # status bar and tool bar
    gui_normal = cv2.WINDOW_GUI_NORMAL  # old fashious way


class MorphShapeEnum(Enum):
    rect = cv2.MORPH_RECT  # a rectangular structuring element: Eij=1
    cross = (
        cv2.MORPH_CROSS
    )  # a cross-shaped structuring element: Eij={10if i=anchor.y or j=anchor.x otherwise
    ellipse = (
        cv2.MORPH_ELLIPSE
    )  # an elliptic structuring element, that is, a filled ellipse inscribed into the rectangle Rect(0, 0, esize.width, 0.esize.height)


class RectanglesIntersectTypes(Enum):
    """
    types of intersection between rectangles
    """

    none = cv2.INTERSECT_NONE
    # No intersection.

    partial = cv2.INTERSECT_PARTIAL
    # There is a partial intersection.

    full = cv2.INTERSECT_FULL
    # One of the rectangle is fully enclosed in the other.


class MorphTypeEnum(Enum):
    erode = cv2.MORPH_ERODE
    dilate = cv2.MORPH_DILATE
    open = cv2.MORPH_OPEN  # dst=open(src,element)=dilate(erode(src,element))
    close = cv2.MORPH_CLOSE  # dst=close(src,element)=erode(dilate(src,element))
    gradient = (
        cv2.MORPH_GRADIENT
    )  # dst=morph_grad(src,element)=dilate(src,element)−erode(src,element)
    tophat = cv2.MORPH_TOPHAT  # dst=tophat(src,element)=src−open(src,element)
    blackhat = cv2.MORPH_BLACKHAT  # dst=blackhat(src,element)=close(src,element)−src
    hitmiss = cv2.MORPH_HITMISS  # Only supported for CV_8UC1 binary images.


class FontEnum(Enum):
    hershey_simplex = cv2.FONT_HERSHEY_SIMPLEX  # normal size sans-serif font

    hershey_plain = cv2.FONT_HERSHEY_PLAIN  # small size sans-serif font

    hershey_duplex = (
        cv2.FONT_HERSHEY_DUPLEX
    )  # normal size sans-serif font (more complex than FONT_HERSHEY_SIMPLEX)

    hershey_complex = cv2.FONT_HERSHEY_COMPLEX  # normal size serif font

    hershey_triplex = (
        cv2.FONT_HERSHEY_TRIPLEX
    )  # normal size serif font (more complex than FONT_HERSHEY_COMPLEX)

    hershey_complex_small = (
        cv2.FONT_HERSHEY_COMPLEX_SMALL
    )  # smaller version of FONT_HERSHEY_COMPLEX

    hershey_script_simplex = cv2.FONT_HERSHEY_SCRIPT_SIMPLEX  # hand-writing style font

    hershey_script_complex = (
        cv2.FONT_HERSHEY_SCRIPT_COMPLEX
    )  # more complex variant of FONT_HERSHEY_SCRIPT_SIMPLEX

    italic = cv2.FONT_ITALIC  # flag for italic font


class ContourRetrievalModeEnum(Enum):
    """

      RETR_EXTERNAL retrieves only the extreme outer contours. It sets hierarchy[i][2]=hierarchy[i][3]=-1 for all the contours.
    RETR_LIST retrieves all of the contours without establishing any hierarchical relationships.
    RETR_CCOMP retrieves all of the contours and organizes them into a two-level hierarchy. At the top level, there are external boundaries of the components. At the second level, there are boundaries of the holes. If there is another contour inside a hole of a connected component, it is still put at the top level.
    RETR_TREE retrieves all of the contours and reconstructs a full hierarchy of nested contours. This full hierarchy is built and shown in the OpenCV contours.c demo.

      :return:
      :rtype:
    """

    external = cv2.RETR_EXTERNAL
    list_all = cv2.RETR_LIST
    ccomp = cv2.RETR_CCOMP
    tree = cv2.RETR_TREE


class ContourApproximationModeEnum(Enum):
    """
    the contour approximation algorithm
    """

    none = cv2.CHAIN_APPROX_NONE
    # stores absolutely all the contour points. That is, any 2 subsequent points (x1,y1) and (x2,y2) of the contour will be either horizontal, vertical or diagonal neighbors, that is, max(abs(x1-x2),abs(y2-y1))==1.

    simple = cv2.CHAIN_APPROX_SIMPLE
    # compresses horizontal, vertical, and diagonal segments and leaves only their end points. For example, an up-right rectangular contour is encoded with 4 points.

    tc89_l1 = cv2.CHAIN_APPROX_TC89_L1
    # applies one of the flavors of the Teh-Chin chain approximation algorithm [244]

    tc89_kcos = cv2.CHAIN_APPROX_TC89_KCOS
    # applies one of the flavors of the Teh-Chin chain approximation algorithm [244]


class MouseEventEnum(Enum):
    move = (
        cv2.EVENT_MOUSEMOVE
    )  # indicates that the mouse pointer has moved over the window.
    left_down = (
        cv2.EVENT_LBUTTONDOWN
    )  # indicates that the left mouse button is pressed.
    left_up = cv2.EVENT_LBUTTONUP  # indicates that left mouse button is released.
    left_double = (
        cv2.EVENT_LBUTTONDBLCLK
    )  # indicates that left mouse button is double clicked.
    right_down = (
        cv2.EVENT_RBUTTONDOWN
    )  # indicates that the right mouse button is pressed.
    right_up = cv2.EVENT_RBUTTONUP  # indicates that right mouse button is released.
    right_double = (
        cv2.EVENT_RBUTTONDBLCLK
    )  # indicates that right mouse button is double clicked.
    middle_down = (
        cv2.EVENT_MBUTTONDOWN
    )  # indicates that the middle mouse button is pressed.
    middle_up = cv2.EVENT_MBUTTONUP  # indicates that middle mouse button is released.
    middle_double = (
        cv2.EVENT_MBUTTONDBLCLK
    )  # indicates that middle mouse button is double clicked.
    wheel_ud = (
        cv2.EVENT_MOUSEWHEEL
    )  # positive and negative values mean forward and backward scrolling, respectively.
    wheel_rl = (
        cv2.EVENT_MOUSEHWHEEL
    )  # positive and negative values mean right and left scrolling, respectively.


class LineTypeEnum(Enum):
    filled = cv2.FILLED
    line4 = cv2.LINE_4  # 4-connected line
    line8 = cv2.LINE_8  # 8-connected line
    anti_aliased = cv2.LINE_AA  # antialiased line


if __name__ == "__main__":
    pass



from enum import Enum

import cv2
from sorcery import assigned_names

from draugr.opencv_utilities.color_space.color import is_singular_channel
from draugr.opencv_utilities.namespaces.color_conversion_enum import (
    ColorConversionEnum,
)

__all__ = ["ToGrayMethodEnum", "to_gray"]


class ToGrayMethodEnum(Enum):
    gray, rgb, hsv, ycrcb, yuv, lab, luv = assigned_names()


def to_gray(
    image,
    *,
    component: int = 0,
    to_gray_method: ToGrayMethodEnum = ToGrayMethodEnum.gray
):
    """
    convert from the default bgr cv2 format to gray, using a single component

    :param image:
    :type image:
    :param component:
    :type component:
    :param to_gray_method:
    :type to_gray_method:
    :return:
    :rtype:
    """
    if not is_singular_channel(image):
        to_gray_method = ToGrayMethodEnum(to_gray_method)
        if to_gray_method == to_gray_method.gray:
            # Weighted sum of the three channels
            # 0.299*R + 0.587*G + 0.114*B
            components = (cv2.cvtColor(image, ColorConversionEnum.bgr2gray.value),)
        elif to_gray_method == ToGrayMethodEnum.rgb:  # Red Green Blue
            # Pick a single channel
            components = cv2.split(
                cv2.cvtColor(image, ColorConversionEnum.bgr2rgb.value)
            )
        elif (
            to_gray_method == ToGrayMethodEnum.hsv
        ):  # Hue ( Dominant Wavelength ) S – Saturation ( Purity / shades of the color ) V – Value ( Intensity )
            # The H Component is very similar in both the images which indicates the color information is intact even under illumination changes.
            # The S component is also very similar in both images.
            # The V Component captures the amount of light falling on it thus it changes due to illumination changes.
            # There is drastic difference between the values of the red piece of outdoor and Indoor image. This is because Hue is represented as a circle and red is at the starting angle. So, it may take values between [300, 360] and again [0, 60].
            components = cv2.split(
                cv2.cvtColor(image, ColorConversionEnum.bgr2hsv.value)
            )
        elif (
            to_gray_method == ToGrayMethodEnum.ycrcb
        ):  # Y – Luma ( Luminance ) Cb – Chroma Blue ( Color ) Cr – Chroma Red ( Color )
            # Y – Luminance or Luma component obtained from RGB after gamma correction.
            # Cr = R – Y ( how far is the red component from Luma ).
            # Cb = B – Y ( how far is the blue component from Luma ).
            # Similar observations as LAB can be made for Intensity and color components with regard to Illumination changes.
            # Perceptual difference between Red and Orange is less even in the outdoor image as compared to LAB.
            # White has undergone change in all 3 components.
            components = cv2.split(
                cv2.cvtColor(image, ColorConversionEnum.bgr2ycrcb.value)
            )
        elif to_gray_method == ToGrayMethodEnum.yuv:
            # Y, a measure of overall brightness or luminance. U and V are computed as scaled differences between Y′ and the B and R values.
            components = cv2.split(
                cv2.cvtColor(image, ColorConversionEnum.bgr2yuv.value)
            )
        elif (
            to_gray_method == to_gray_method.lab
        ):  # L – Lightness ( Brightness ) a – Green-Magenta ( Green - Red ) b – Blue-Yellow ( Blue - Red )
            # Illumination has mostly affected the L component.
            # The A and B components which contain the color information did not undergo massive changes.
            # The respective values of Green, Orange and Red ( which are the extremes of the A Component ) has not changed in the B Component and similarly the respective values of Blue and Yellow ( which are the extremes of the B Component ) has not changed in the A component.
            components = cv2.split(
                cv2.cvtColor(image, ColorConversionEnum.bgr2lab.value)
            )
        elif to_gray_method == to_gray_method.luv:
            # LUV decouple the "color" (chromaticity, the UV part) and "lightness" (luminance, the L part)
            components = cv2.split(
                cv2.cvtColor(image, ColorConversionEnum.bgr2luv.value)
            )
        else:
            raise NotImplementedError
        return components[component]
    return image



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

__author__ = "Christian Heider Nielsen"
__doc__ = r"""

           Created on 03/08/2020
           """

__all__ = ["GtkSettings"]

from sys import stderr

import gi  # PyGObject lib

#  Ubuntu / Debian
## system
### sudo apt install python3-gi python3-gi-cairo gir1.2-gtk-3.0

## conda
### conda install -c conda-forge pygobject
### conda install -c conda-forge gtk3

## pip
### sudo apt install libgirepository1.0-dev gcc libcairo2-dev pkg-config python3-dev gir1.2-gtk-3.0
### pip3 install pycairo
### pip3 install PyGObject

gi.require_version("Gtk", "3.0")
gi.require_version("Gdk", "3.0")

from warg import SingletonMeta


class GtkSettings(metaclass=SingletonMeta):
    """ """

    def __init__(self):
        try:
            from gi.repository import Gtk, Gdk  # , GObject, Gio, GLib
        except (ImportError, ImportWarning):
            stderr.write("Could not import GTK. Please install it.")

        # a = Gio.Settings("org.gnome.gedit.preferences.editor")
        # print(a.get_string("wrap-mode"))
        # settings = Gio.Settings('org.my.font')
        # face = settings.get_string('default-face')
        # size = settings.get_double('default-size')

        self.settings = Gtk.Settings.get_default()
        # self.settings = Gtk.Settings.get_for_screen(Gdk.get_default_root_window().get_screen())

    def __getitem__(self, item):
        return self.settings.get_property(item)

    def __setitem__(self, key, value):
        self.settings.set_property(key, value)

    @property
    def all_settings(self):
        """ """
        for i in self.settings.list_properties():  # getting all existing properties
            yield i

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        return 0  # self.settings


if __name__ == "__main__":

    def abvdfj() -> None:
        """
        :rtype: None
        """
        a = GtkSettings()

        def ofgof():
            """ """
            for b in a.all_settings:
                print(b)

        ofgof()

        print(a["gtk-dialogs-use-header"])
        # print(a['gtk-dialogs-use-header1'])

    abvdfj()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
from abc import ABC, abstractmethod
from enum import Enum
from typing import Union

import numpy
import torch
from PIL import Image

__author__ = "Christian Heider Nielsen"
__doc__ = r"""

           Created on 09/10/2019
           """
__all__ = ["VideoWriterMixin", "VideoInputDimsEnum"]

from warg import Number


class VideoInputDimsEnum(Enum):
    """
    Input mode"""

    thw = "THW"
    tchw = "TCHW"
    thwc = "THWC"
    ntchw = "NTCHW"
    nthwc = "NTHWC"
    tnchw = "TNCHW"
    tnhwc = "TNHWC"


class VideoWriterMixin(ABC):
    """
    Writer subclass that provides an interface for 'writing' video clips"""

    @abstractmethod
    def video(
        self,
        tag: str,
        data: Union[numpy.ndarray, torch.Tensor, Image.Image],
        step,
        frame_rate: Number = 30,
        input_dims: VideoInputDimsEnum = VideoInputDimsEnum.ntchw,
        **kwargs
    ) -> None:
        """

            :param frame_rate:
            :param input_dims:
        :param tag:
        :type tag:
        :param data:
        :type data:
        :param step:
        :type step:
        :param kwargs:
        :type kwargs:"""
        raise NotImplementedError



import os
import subprocess
from pathlib import Path
from typing import Optional

from apppath import ensure_existence
from draugr.ffmpeg_utilities.extract import FORMAT_LIST, AUDIO_FORMAT
from warg import Number, identity

__all__ = ["merge_video"]


def get_short_name(vid_dir) -> str:
    for file_ in vid_dir.iterdir():
        return file_.stem.split("-")[0]
        # if os.path.splitext(file_)[-1].lower() in FORMAT_LIST:
        #    return os.path.splitext(file_)[0]


def get_frame_format(frames_dir) -> str:
    for file_ in os.listdir(frames_dir):
        if os.path.splitext(file_)[-1].lower() in [".jpg", ".png"]:
            return os.path.splitext(file_)[-1].lower()


def merge_video(
    frames_dir: Path,
    merge_audio: bool = False,
    audio_dir: Optional[Path] = None,
    merge_dir: Optional[Path] = None,
    merge_rate: Number = 25,
    ffmpeg_path: Path = "ffmpeg",
):
    postfix = ""  # "_00"
    vid_dir = frames_dir.parent
    if audio_dir is None:
        audio_dir = vid_dir / "audio"

    if merge_dir is None:
        merge_dir = ensure_existence(vid_dir / "merge", sanitisation_func=identity)

    shortname = get_short_name(frames_dir)

    # a = f"{shortname}-%05d{postfix}{get_frame_format(frames_dir)}"  # 05d is for 5 digits in ids
    # a = "brandt.mp4-%*_00.png"
    a = f"%05d{get_frame_format(frames_dir)}"
    temp_out = str(merge_dir / "temp.mp4")
    subprocess.call(
        [
            str(ffmpeg_path),
            "-r",
            str(merge_rate),
            "-i",
            str(frames_dir / a),
            "-y",
            "-c:v",
            "libx264",
            "-vf",
            "fps=25,format=yuv420p",
            temp_out,
        ]
    )

    a = []
    if merge_audio:
        sound_dir = audio_dir / (shortname + AUDIO_FORMAT)
        a.extend(["-i", str(sound_dir)])

    while (merge_dir / f"{shortname}out.mp4").exists():
        shortname = shortname + "_1"

    subprocess.call(
        [
            str(ffmpeg_path),
            "-i",
            temp_out,
            *a,
            "-vcodec",
            "copy",
            "-acodec",
            "copy",
            "-y",
            str(merge_dir / shortname),
        ]
    )


if __name__ == "__main__":
    merge_video(
        # Path.home() / "Downloads" / "brandt" / "frames",
        # Path(r"G:/") / "Mit drev" / "upscaled" / "restored_faces",
        Path(r"G:/") / "Mit drev" / "upscaled" / "restored_imgs",
        ffmpeg_path=Path.home()
        / "OneDrive - Alexandra Instituttet"
        / "Applications"
        / "ffmpeg-5.0-essentials_build"
        / "bin"
        / "ffmpeg.exe",
    )



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

__author__ = "Christian Heider Nielsen"
__doc__ = r"""
          Fission variants of MLPs

           Created on 24/02/2020
           """

from typing import Sequence, Tuple

import torch
from torch import nn


__all__ = ["DisjunctMLP"]

from draugr.torch_utilities.architectures.mlp import MLP


class DisjunctMLP(MLP):
    """ """

    def __init__(
        self,
        output_shape: Sequence = (2,),
        disjunction_size=256,
        subnet_size=128,
        hidden_layer_activation=nn.ReLU(),
        **kwargs
    ):
        super().__init__(
            output_shape=(disjunction_size,),
            hidden_layer_activation=hidden_layer_activation,
            output_activation=nn.Identity(),
            **kwargs
        )

        self.subnet_1 = torch.nn.Sequential(
            torch.nn.Linear(disjunction_size, subnet_size),
            hidden_layer_activation,
            torch.nn.Linear(subnet_size, output_shape[-1]),
        )

        self.subnet_2 = torch.nn.Sequential(
            torch.nn.Linear(disjunction_size, subnet_size),
            hidden_layer_activation,
            torch.nn.Linear(subnet_size, 1),
        )

    def forward(self, *act, **kwargs) -> Tuple[torch.tensor, torch.tensor]:
        """

        :param act:
        :type act:
        :param kwargs:
        :type kwargs:
        :return:
        :rtype:
        """
        x = super().forward(*act, **kwargs)
        return self.subnet_1(x), self.subnet_2(x)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

__author__ = "Christian Heider Nielsen"
__doc__ = r"""

           Created on 18-01-2021
           """

import numpy

from warg import next_pow_2

__all__ = ["zero_pad_to_power_2", "zero_pad_to"]


def zero_pad_to(signal: numpy.ndarray, length: int) -> numpy.ndarray:
    """ """
    return numpy.pad(
        signal, (0, length - len(signal)), "constant", constant_values=(0, 0)
    )


def zero_pad_to_power_2(signal: numpy.ndarray) -> numpy.ndarray:
    """ """
    return zero_pad_to(signal, next_pow_2(len(signal)))


if __name__ == "__main__":
    aasd = numpy.arange(8 + 1)
    print(aasd, aasd.shape)
    padded = zero_pad_to_power_2(aasd)
    print(padded, padded.shape)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

__author__ = "Christian Heider Nielsen"
__doc__ = r"""

           Created on 15/06/2020
           """

from pathlib import Path
from typing import Iterable, Optional, Sequence, Tuple, Union

from cycler import Cycler
from matplotlib import pyplot

from warg import AlsoDecorator, passes_kws_to

__all__ = [
    "FigureSession",
    "SubplotSession",
    "MonoChromeStyleSession",
    "NoOutlineSession",
    "OutlineSession",
    "StyleSession",
]

from draugr.visualisation.matplotlib_utilities.matplotlib_utilities import (
    monochrome_line_cycler,
)
from draugr.visualisation.matplotlib_utilities.quirks import fix_edge_gridlines
from draugr.visualisation.matplotlib_utilities.styles.cyclers import (
    color_cycler,
    line_cycler,
)
import subprocess


class FigureSession(AlsoDecorator):
    """ """

    @passes_kws_to(pyplot.figure)
    def __init__(self, **kws):
        self.fig = pyplot.figure(**kws)

    def __enter__(self) -> pyplot.Figure:
        return self.fig

    def __exit__(self, exc_type, exc_val, exc_tb):
        pyplot.cla()
        pyplot.close(self.fig)
        pyplot.clf()


class SubplotSession(AlsoDecorator):
    """ """

    @passes_kws_to(pyplot.subplots)
    def __init__(self, return_self: bool = False, **kws):
        self.fig, axs = pyplot.subplots(**kws)
        if not isinstance(axs, Iterable):
            axs = (axs,)
        self.axs = axs
        self.return_self = return_self

    def __enter__(
        self,
    ) -> Union["SubplotSession", Tuple[pyplot.Figure, Sequence[pyplot.Axes]]]:
        if self.return_self:
            return self
        return self.fig, self.axs

    def __exit__(self, exc_type, exc_val, exc_tb):
        # pyplot.cla()
        # pyplot.clf()
        self.fig.clear()
        pyplot.close(self.fig)


class StyleSession(AlsoDecorator):
    """ """

    def __init__(
        self,
        style_path: Path = Path(__file__).parent / "styles" / "publish_color.mplstyle",
        prop_cycler: Optional[Cycler] = line_cycler + color_cycler,
    ):
        """
        Set styling for context

        :param style_path: path to style file
        :param prop_cycler: overwrite prop_cycler with a programmatically defined one
        """
        self.ctx = pyplot.style.context(style_path)
        self.prop_cycler = prop_cycler

    def __enter__(self) -> pyplot.Figure:
        a = self.ctx.__enter__()
        if pyplot.rcParams["text.usetex"]:
            try:
                report = subprocess.check_output("latex -v", stderr=subprocess.STDOUT)
            except FileNotFoundError as exc:
                msg = f'No tex: {"latex"}'
                # raise RuntimeError(msg)
                print(f"{msg}, disabling")
                pyplot.rcParams.update({"text.usetex": False})
        if self.prop_cycler:
            pyplot.rcParams.update({"axes.prop_cycle": self.prop_cycler})
        return a

    def __exit__(self, exc_type, exc_val, exc_tb):
        fix_edge_gridlines(pyplot.gca())
        # pyplot.tight_layout()
        # auto_post_hatch(pyplot.gca(),hatch_cycler)
        self.ctx.__exit__(exc_type, exc_val, exc_tb)


class MonoChromeStyleSession(StyleSession):
    """ """

    def __init__(
        self,
        style_path=Path(__file__).parent / "styles" / "monochrome.mplstyle",
        prop_cycler: Optional[Cycler] = monochrome_line_cycler,
    ):
        """

        :param style_path:
        :param prop_cycler:
        """
        super().__init__(style_path, prop_cycler)


class NoOutlineSession(AlsoDecorator):
    """ """

    def __init__(self):
        self._rcParams_copy = pyplot.rcParams.copy()

    def __enter__(self) -> bool:
        pyplot.rcParams.update(
            {
                "patch.edgecolor": "none",
                "patch.force_edgecolor": False,
                "patch.linewidth": 0,
            }
        )
        return True

    def __exit__(self, exc_type, exc_val, exc_tb):
        pyplot.rcParams = self._rcParams_copy


class OutlineSession(AlsoDecorator):
    """ """

    def __init__(self):
        self._rcParams_copy = pyplot.rcParams.copy()

    def __enter__(self) -> bool:
        pyplot.rcParams.update(
            {
                "patch.edgecolor": "k",
                "patch.force_edgecolor": True,
                "patch.linewidth": 1.0,
            }
        )
        return True

    def __exit__(self, exc_type, exc_val, exc_tb):
        pyplot.rcParams = self._rcParams_copy


if __name__ == "__main__":

    def deiajsd() -> None:
        """
        :rtype: None
        """
        for _ in range(100):
            with SubplotSession() as a:
                fig, (ax1,) = a
                ax1.set_ylabel("test")

    def asiuhdsada() -> None:
        """
        :rtype: None
        """
        with MonoChromeStyleSession():
            fig, ax = pyplot.subplots(1, 1)
            for x in range(3):
                import numpy

                ax.plot(numpy.random.rand(10), label=f"{x}")

        from matplotlib.pyplot import legend

        legend()
        pyplot.show()

    def no_border_boxplot() -> None:
        """
        :rtype: None
        """
        with NoOutlineSession():
            import numpy

            num = 5
            pyplot.bar(range(num), numpy.random.rand(num))
        pyplot.show()

    def border_boxplot() -> None:
        """
        :rtype: None
        """
        with OutlineSession():
            import numpy

            num = 5
            pyplot.bar(range(num), numpy.random.rand(num))
            pyplot.show()

    # deiajsd()
    no_border_boxplot()
    # border_boxplot()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
import math

import numpy
from matplotlib.gridspec import GridSpec

from draugr.drawers.mpl_drawers.mpldrawer import MplDrawer

__author__ = "Christian Heider Nielsen"
__doc__ = """
Created on 27/04/2019

@author: cnheider
"""

__all__ = ["FastFourierTransformSpectrogramPlot"]

from matplotlib import pyplot

from draugr.tqdm_utilities import progress_bar

FLOAT_EPS = numpy.finfo(float).eps


class FastFourierTransformSpectrogramPlot(MplDrawer):
    """
    TODO: CENTER Align fft maybe, to mimick librosa stft
    Short Time Fourier Transform (STFT), with step size of 1 and window lenght of n_fft, and no window function ( TODO: Hanning Smoothing)"""

    def __init__(
        self,
        n_fft: int = 64,
        sampling_rate=int(1.0 / 0.0005),
        buffer_size_sec: float = 1.0,
        title: str = "",
        vertical: bool = True,
        reverse: bool = False,
        figure_size=(9, 9),
        cmap="viridis",
        render: bool = True,
        **kwargs
    ):
        """

        :param n_fft:
        :type n_fft:
        :param sampling_rate:
        :type sampling_rate:
        :param buffer_size_sec:
        :type buffer_size_sec:
        :param title:
        :type title:
        :param vertical:
        :type vertical:
        :param reverse:
        :type reverse:
        :param placement:
        :type placement:
        :param fig_size:
        :type fig_size:
        :param render:
        :type render:"""
        super().__init__(render=render, figure_size=figure_size, **kwargs)

        if not render:
            return

        self.fig = pyplot.figure(figsize=figure_size)

        self.n_fft = n_fft
        self.n_positive_fft = (self.n_fft + 1) // 2
        self.sampling_rate = sampling_rate
        if buffer_size_sec is not None:
            self.buffer_size_sec = buffer_size_sec
        else:
            self.buffer_size_sec = self.n_fft / sampling_rate * 2

        # self.window_function = window_function(self.n_fft)
        self.buffer_array_size = int(sampling_rate * self.buffer_size_sec)
        assert self.buffer_array_size >= self.n_fft
        time_s = numpy.linspace(
            0, self.buffer_size_sec, self.buffer_array_size, endpoint=False
        )
        raw_array = numpy.zeros(self.buffer_array_size)
        zeroes_img = numpy.zeros((self.n_positive_fft, self.buffer_array_size - n_fft))
        self.zeroes_padding = numpy.zeros((self.n_positive_fft, n_fft))

        gs = GridSpec(3, 2, width_ratios=[100, 2])
        (
            self.raw_ax,
            _,
            self.angle_ax,
            self.angle_cbar_ax,
            self.spec_ax,
            self.spec_cbar_ax,
        ) = [pyplot.subplot(gs[i]) for i in range(6)]

        (self.raw_line2d,) = self.raw_ax.plot(time_s, raw_array)
        self.raw_ax.set_xlim([time_s[0], time_s[-1]])
        self.raw_ax.set_ylabel("Signal [Magnitude]")

        max_freq = numpy.max(numpy.fft.fftfreq(self.n_fft, 1 / sampling_rate))
        self.dft_angle_img = self.angle_ax.imshow(
            zeroes_img,
            vmin=-math.pi,
            vmax=math.pi,
            interpolation="hanning",
            aspect="auto",
            extent=[time_s[0], time_s[-1], max_freq, 0],
            cmap=cmap,
        )
        self.angle_ax.set_ylabel("Phase [Hz]")
        _ = self.fig.colorbar(self.dft_angle_img, cax=self.angle_cbar_ax)
        self.angle_cbar_ax.set_ylabel("Angle (Radians)", rotation=90)

        self.dft_mag_img = self.spec_ax.imshow(
            zeroes_img,
            vmin=0,
            vmax=1,
            interpolation="hanning",
            aspect="auto",
            extent=[time_s[0], time_s[-1], max_freq, 0],
            cmap=cmap,
        )
        self.spec_ax.set_ylabel("Frequency [Hz]")
        self.spec_ax.set_xlabel("Time [Sec]")
        _ = self.fig.colorbar(self.dft_mag_img, cax=self.spec_cbar_ax)
        self.spec_cbar_ax.set_ylabel("Magnitude (dB)", rotation=90)

        self.vertical = vertical
        self.reverse = reverse

        pyplot.xlim(time_s[0], time_s[-1])

        pyplot.title(title)
        pyplot.tight_layout()

    def _draw(self, signal_sample: float, delta: float = 1 / 120) -> None:
        """

        :param signal_sample:
        :param delta: 1 / 60 for 60fps
        :return:"""
        y_data = self.raw_line2d.get_ydata()

        if not self.reverse:
            y_data = numpy.hstack((y_data[1:], signal_sample))
            y_data_view = y_data[-self.n_fft :]
        else:
            y_data = numpy.hstack((signal_sample, y_data[:-1]))
            y_data_view = y_data[: self.n_fft]

        self.raw_line2d.set_ydata(y_data)
        cur_lim = self.raw_ax.get_ylim()
        self.raw_ax.set_ylim(
            [min(cur_lim[0], signal_sample), max(cur_lim[1], signal_sample)]
        )

        # if self.window_function is not None:
        #   y_data_view *= self.window_function

        f_coef = numpy.fft.fft(y_data_view, n=self.n_fft)[
            : self.n_positive_fft
        ].reshape(
            -1, 1
        )  # Only select the positive frequencies

        phase_data = self.dft_angle_img.get_array()
        new_phase = numpy.angle(f_coef)
        # new_phase = f_coef.imag

        if not self.reverse:
            phase_data = numpy.concatenate(
                (phase_data[:, 1 : -self.n_fft], new_phase, self.zeroes_padding),
                axis=-1,
            )
        else:
            phase_data = numpy.concatenate(
                (self.zeroes_padding, new_phase, phase_data[:, self.n_fft : -1]),
                axis=-1,
            )

        self.dft_angle_img.set_array(phase_data)

        magnitude_data = self.dft_mag_img.get_array()
        new_mag = 10 * numpy.log10(
            (
                numpy.abs(f_coef)
                # f_coef.real
                ** 2
            )
            + FLOAT_EPS
        )  # db
        if not self.reverse:
            magnitude_data = numpy.concatenate(
                (magnitude_data[:, 1 : -self.n_fft], new_mag, self.zeroes_padding),
                axis=-1,
            )
        else:
            magnitude_data = numpy.concatenate(
                (self.zeroes_padding, new_mag, magnitude_data[:, self.n_fft : -1]),
                axis=-1,
            )

        self.dft_mag_img.set_clim(
            vmin=numpy.min(magnitude_data), vmax=numpy.max(magnitude_data)
        )
        self.dft_mag_img.set_array(magnitude_data)
        # self.spec_cbar_ax.set_ylabel("Magnitude (Linear)", rotation=90)
        self.spec_cbar_ax.set_ylabel("Magnitude (dB)", rotation=90)


if __name__ == "__main__":

    def a() -> None:
        """
        :rtype: None
        """
        duration_sec = 4
        mul = 1000
        sampling_Hz = 44.1
        sampling_rate = int(sampling_Hz * mul)  # Hz
        delta = 1 / sampling_rate
        n_fft = 128  # 1024
        s = FastFourierTransformSpectrogramPlot(
            n_fft=n_fft, sampling_rate=sampling_rate, buffer_size_sec=delta * n_fft * 4
        )
        for t in progress_bar(numpy.arange(0, duration_sec, delta)):
            ts = 2 * numpy.pi * t
            s1 = numpy.sin(ts * 1 * sampling_Hz / 2**4 * mul)
            s2 = numpy.sin(ts * 3 * sampling_Hz / 2**3 * mul + 0.33 * numpy.pi)
            s3 = numpy.sin(ts * 5 * sampling_Hz / 2**2 * mul + 0.66 * numpy.pi)
            signal = s1 + s2 + s3
            signal /= 3
            # signal += (numpy.random.random() - 0.5) * 2 * 1 / 2  # Noise
            s.draw(signal, delta=delta)

    a()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

__author__ = "Christian Heider Nielsen"
__doc__ = r"""

           Created on 14-03-2021
           """

from draugr.writers import Writer


class SqliteWriter(Writer):
    pass  # TODO:



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

__author__ = "heider"
__doc__ = r"""

           """

__all__ = []

import ctypes
import ctypes.util

try:  # macOS Big Sur+ use "a built-in dynamic linker cache of all system-provided libraries"

    appkit = ctypes.cdll.LoadLibrary("AppKit.framework/AppKit")
    objc = ctypes.cdll.LoadLibrary("libobjc.dylib")
except OSError:  # revert to full path for older OS versions and hardened programs

    appkit = ctypes.cdll.LoadLibrary(ctypes.util.find_library("AppKit"))
    objc = ctypes.cdll.LoadLibrary(ctypes.util.find_library("objc"))

void_p = ctypes.c_void_p
ull = ctypes.c_uint64

objc.objc_getClass.restype = void_p
objc.sel_registerName.restype = void_p

# See https://docs.python.org/3/library/ctypes.html#function-prototypes for arguments description
MSGPROTOTYPE = ctypes.CFUNCTYPE(void_p, void_p, void_p, void_p)
msg = MSGPROTOTYPE(
    ("objc_msgSend", objc), ((1, "", None), (1, "", None), (1, "", None))
)


def _utf8(s):
    if not isinstance(s, bytes):
        s = s.encode("utf8")
    return s


def n(name):
    return objc.sel_registerName(_utf8(name))


def C(classname):
    return objc.objc_getClass(_utf8(classname))


def is_dark_mode_active():
    NSAutoreleasePool = objc.objc_getClass("NSAutoreleasePool")
    pool = msg(NSAutoreleasePool, n("alloc"))
    pool = msg(pool, n("init"))

    NSUserDefaults = C("NSUserDefaults")
    stdUserDef = msg(NSUserDefaults, n("standardUserDefaults"))

    NSString = C("NSString")

    key = msg(NSString, n("stringWithUTF8String:"), _utf8("AppleInterfaceStyle"))
    appearanceNS = msg(stdUserDef, n("stringForKey:"), void_p(key))
    appearanceC = msg(appearanceNS, n("UTF8String"))

    if appearanceC is not None:
        out = ctypes.string_at(appearanceC)
    else:
        out = None

    msg(pool, n("release"))

    if out is not None:
        if out.decode("utf-8") == "Dark":
            return True

    return False


if __name__ == "__main__":
    print(is_dark_mode_active())



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

__author__ = "Christian Heider Nielsen"
__doc__ = r"""

           Created on 14-03-2021
           """

from typing import Tuple

import numpy

from warg import Number

__all__ = ["circular_mask"]


def circular_mask(
    h: Number, w: Number, center: Tuple[Number, Number] = None, radius: Number = None
) -> numpy.ndarray:
    """

    :param h:
    :param w:
    :param center:
    :param radius:
    :return:"""
    if center is None:  # use the middle of the image
        center = (int(w / 2), int(h / 2))
    if radius is None:  # use the smallest distance between the center and image walls
        radius = min(center[0], center[1], w - center[0], h - center[1])

    y, x = numpy.ogrid[:h, :w]
    dist_from_center = numpy.sqrt((x - center[0]) ** 2 + (y - center[1]) ** 2)

    return dist_from_center <= radius


if __name__ == "__main__":
    print(circular_mask(9, 9))



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

__author__ = "Christian Heider Nielsen"
__doc__ = r"""
"""

from pathlib import Path

with open(Path(__file__).parent / "README.md", "r") as this_init_file:
    __doc__ += this_init_file.read()

from .display_3d_depth_image import *
from .display_depth_image import *
from .replay_frames import *
from .multiple import *



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

__author__ = "Christian Heider Nielsen"
__doc__ = r"""

           Created on 17-05-2021
           """

import psutil

PAGING_BUFFER_SIZE = 0.2 * psutil.virtual_memory().total

__all__ = ["num_instance_no_paging"]


def num_instance_no_paging(expected_size_mb: int = 1024) -> int:
    """

    :param expected_size_mb:
    :return:
    """
    return int(
        (psutil.virtual_memory().available - PAGING_BUFFER_SIZE)
        / expected_size_mb
        * 1e-6
    )


if __name__ == "__main__":
    print(num_instance_no_paging())



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

__author__ = "Christian Heider Nielsen"
__doc__ = r"""

           Created on 12-05-2021
           """

__all__ = [
    "min_interval_wrapper",
    "min_interval_wrapper_global",
    "max_frequency",
    "wrap_args",
]

from collections import namedtuple
from typing import MutableMapping, Tuple, Any

import wrapt


def min_interval_wrapper(f: callable, min_interval: int = 100) -> callable:
    """
    to ensure that a function is now being called more often than max_freq, TODO: use proper naming for the interval
    :param f:
    :param min_interval:
    :return:
    """

    def s(
        last_exec: int, *, step_i: int, verbose: bool = False, **kwargs: MutableMapping
    ) -> int:
        """

        :param last_exec:
        :param step_i:
        :param verbose:
        :param kwargs:
        :return:
        """
        if verbose:
            print(f"{f, last_exec, step_i, min_interval}")
        if step_i - last_exec >= min_interval:
            f(step_i=step_i, verbose=verbose, **kwargs)
            return step_i
        return last_exec

    return s


max_frequency_wrapper = min_interval_wrapper

_GLOBAL_COUNTERS = {}


def min_interval_wrapper_global(f: callable, min_interval: int = 100) -> callable:
    """
    to ensure that a function is now being called more often than max_freq, TODO: use proper naming for the interval
    :param f:
    :param min_interval:
    :return:
    """

    _GLOBAL_COUNTERS[f] = 0

    def s(*, step_i: int, verbose: bool = False, **kwargs: MutableMapping) -> None:
        """

        :param step_i:
        :param verbose:
        :param kwargs:
        """
        if verbose:
            print(f"{f, _GLOBAL_COUNTERS[f], step_i, min_interval}")
        if step_i - _GLOBAL_COUNTERS[f] >= min_interval:
            _GLOBAL_COUNTERS[f] = step_i
            f(step_i=step_i, verbose=verbose, **kwargs)

    return s


max_frequency_wrapper_global = min_interval_wrapper_global


def max_frequency(key: Any, min_interval: int = 100, verbose: bool = False) -> callable:
    """
    initially returns recallable func later bools
        :param f:
    :param min_interval:
    :return:
    """

    if key in _GLOBAL_COUNTERS:
        if verbose:
            print(f"{key, _GLOBAL_COUNTERS[key], min_interval}")
        if _GLOBAL_COUNTERS[key] >= min_interval:
            _GLOBAL_COUNTERS[key] = 1
            return True
        _GLOBAL_COUNTERS[key] += 1
        return False
    else:
        _GLOBAL_COUNTERS[key] = min_interval

        def s() -> bool:
            """

            :param step_i:
            :param verbose:
            :param kwargs:
            """
            if verbose:
                print(f"{key, _GLOBAL_COUNTERS[key], min_interval}")
            if _GLOBAL_COUNTERS[key] >= min_interval:
                _GLOBAL_COUNTERS[key] = 1
                return True
            _GLOBAL_COUNTERS[key] += 1
            return False

        return s


def wrap_args(n_tuple: namedtuple):
    """

    :param n_tuple:
    :type n_tuple:
    :return:
    :rtype:"""

    @wrapt.decorator(adapter=n_tuple)
    def wrapper(wrapped, instance, args, kwargs):
        """

        :param wrapped:
        :type wrapped:
        :param instance:
        :type instance:
        :param args:
        :type args:
        :param kwargs:
        :type kwargs:
        :return:
        :rtype:"""
        if isinstance(args[0], n_tuple):
            n = args[0]
        else:
            n = n_tuple(*args, **kwargs)
        return wrapped(n)

    return wrapper


def str_to_bool(s: str, preds: Tuple[str, ...] = ("true", "1")) -> bool:
    """


    :param preds:
    :param s:
    :return:"""
    return s.lower() in preds


str2bool = str_to_bool

if __name__ == "__main__":

    c = namedtuple("C", ("a", "b"))

    @wrap_args(c)
    def add(v):
        """

        :param v:
        :type v:
        :return:
        :rtype:"""
        return v.a + v.b

    def add2(a, b):
        """

        :param a:
        :type a:
        :param b:
        :type b:
        :return:
        :rtype:"""
        return a + b

    h = add(2, 2)
    print(h)

    j = add(c(1, 4))
    print(j)

    wq = add2(2, 4)
    print(wq)

    wc = add2(*c(4, 3))
    print(wc)

    def a(step_i, **kwargs):
        """

        :param step_i:
        :param kwargs:
        """
        print(step_i)

    def uhsud() -> None:
        """
        :rtype: None
        """
        b = min_interval_wrapper(a)
        c = 0
        for i in range(1000 + 1):
            c = b(c, step_i=i)

    def uhsud23() -> None:
        """
        :rtype: None
        """
        from random import random

        b = min_interval_wrapper_global(a)

        for i in range(1000 + 1):
            if random() > 0.8:
                b(step_i=i)

    def uhsud123() -> None:
        """
        :rtype: None
        """
        from random import random

        b = min_interval_wrapper_global(a, 0)

        for i in range(1000 + 1):
            if random() > 0.8 or True:
                b(step_i=i)

    # uhsud123()

    def iuhasd():
        from random import random

        a = 0
        for i in range(1000 + 1):
            if True:
                if max_frequency("key1", 100):
                    a += 1
                    print(i, a)

    def iuhasd2():
        from random import random

        f = max_frequency("key2", 100)
        a = 0
        for i in range(1000 + 1):
            if f():
                a += 1
                print(i, a)

    iuhasd()
    iuhasd2()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

__author__ = "Christian Heider Nielsen"
__doc__ = r"""

           Created on 30/03/2020
           """

from typing import Tuple

COLOR_RGB = Tuple[int, int, int]
COLOR_RGBA = Tuple[int, int, int, int]
COLOR_INT = Tuple[int, ...]

__all__ = ["rgb", "rgba", "RGB", "RGBA", "color_to_str", "color_from_str"]


def rgb(r: int, g: int, b: int) -> COLOR_RGB:
    """

    :param r:
    :type r:
    :param g:
    :type g:
    :param b:
    :type b:
    :return:
    :rtype:"""
    assert 0 <= r <= 255 and 0 <= g <= 255 and 0 <= b <= 255
    return r, g, b


RGB = rgb


def rgba(r: int, g: int, b: int, a: int) -> COLOR_RGBA:
    """

    :param r:
    :type r:
    :param g:
    :type g:
    :param b:
    :type b:
    :param a:
    :type a:
    :return:
    :rtype:"""
    assert 0 <= r <= 255 and 0 <= g <= 255 and 0 <= b <= 255 and 0 <= a <= 255
    return r, g, b, a


RGBA = rgba


def color_from_str(s: str, seperator: str = " ") -> COLOR_INT:
    """

    :param s:
    :type s:
    :param seperator:
    :type seperator:
    :return:
    :rtype:"""
    components = s.split(seperator)
    n = len(components)
    if n == 3:
        return RGB(*[int(c) for c in components])
    elif n == 4:
        return RGBA(*[int(c) for c in components])
    raise NotImplementedError("Color space not recognised")


def color_to_str(t: COLOR_INT, seperator: str = " ") -> str:
    """

    :param t:
    :type t:
    :param seperator:
    :type seperator:
    :return:
    :rtype:"""
    return seperator.join([str(c) for c in t])


if __name__ == "__main__":

    def main() -> None:
        """
        :rtype: None
        """
        a = RGB(1, 50, 100)
        b = color_to_str(a)
        c = color_from_str(b)
        print(c)

    main()



from enum import Enum

import cv2
import numpy
from sorcery import assigned_names

__all__ = ["threshold_channel", "ThresholdMethodEnum", "hsv_min_max_clip_mask"]

from draugr.opencv_utilities.namespaces.flags import ThresholdTypeFlag
from draugr.opencv_utilities.namespaces.color_conversion_enum import ColorConversionEnum


class ThresholdMethodEnum(Enum):
    simple, adaptive = assigned_names()


def threshold_channel(
    gray: numpy.ndarray,
    method: ThresholdMethodEnum = ThresholdMethodEnum.simple,
    **kwargs
) -> numpy.ndarray:
    method = ThresholdMethodEnum(method)

    if method == ThresholdMethodEnum.simple:
        return cv2.threshold(
            gray,
            thresh=kwargs.get("thresh", 120),
            maxval=kwargs.get("maxval", 255),
            type=ThresholdTypeFlag.otsu.value
            + ThresholdTypeFlag.binary.value,  # +ThresholdTypeFlag.to_zero.value,
        )[1]
    elif method == ThresholdMethodEnum.adaptive:
        return cv2.adaptiveThreshold(
            gray,
            maxValue=kwargs.get("maxValue", 255),
            adaptiveMethod=cv2.ADAPTIVE_THRESH_GAUSSIAN_C,
            thresholdType=cv2.THRESH_BINARY,
            blockSize=11,
            C=2,
        )
    raise NotImplementedError


def hsv_min_max_clip_mask(
    input_image: numpy.ndarray,
    lower_bound: numpy.ndarray = numpy.array([0, 0, 0]),
    upper_bound: numpy.ndarray = numpy.array(
        [179, 255, 255]
    ),  # Hue is from 0-179 for Opencv
) -> numpy.ndarray:
    return cv2.inRange(
        cv2.cvtColor(input_image, ColorConversionEnum.bgr2hsv.value),
        lowerb=lower_bound,
        upperb=upper_bound,
    )



import os
import subprocess
from pathlib import Path
from typing import Optional

from apppath import ensure_existence
from draugr.ffmpeg_utilities.extract import FORMAT_LIST, AUDIO_FORMAT
from warg import Number, identity


__all__ = ["sequencify_files"]


def sequencify_files(frames_dir: Path):

    for i, f in enumerate(sorted(frames_dir.iterdir(), key=lambda x: x.name)):
        if f.is_file():
            suffix = {f.suffix}
            # suffix = ".jpg"
            f.rename(frames_dir / f"{i:05d}{suffix}")


if __name__ == "__main__":
    sequencify_files(
        Path(r"G:/") / "Mit drev" / "upscaled" / "restored_imgs",
    )



from pathlib import Path
from typing import Iterable

import cv2
import numpy

from draugr.opencv_utilities.color_space.threshold import hsv_min_max_clip_mask
from draugr.opencv_utilities.windows.image import show_image
from draugr.opencv_utilities.windows.elements import add_trackbar


def interative_hsv_color_picker(ps: Iterable, waitTime=33, verbose=False):

    show_image(numpy.zeros((600, 600)), "image")

    add_trackbar(
        "image", "HMin", default=0, max_val=179
    )  # Hue is from 0-179 for Opencv
    add_trackbar("image", "HMax", default=179, max_val=179)

    add_trackbar("image", "SMin", default=0)
    add_trackbar("image", "SMax", default=255)

    add_trackbar("image", "VMin", default=0)
    add_trackbar("image", "VMax", default=255)

    # Initialize to check if HSV min/max value changes
    h_min = s_min = v_min = h_max = s_max = v_max = 0
    ph_min = ps_min = pv_min = ph_max = ps_max = pv_max = 0

    for p in ps:
        if not p.exists():
            continue

        img = cv2.imread(str(p))
        output = img
        while 1:

            h_min = cv2.getTrackbarPos("HMin", "image")
            h_max = cv2.getTrackbarPos("HMax", "image")

            s_min = cv2.getTrackbarPos("SMin", "image")
            s_max = cv2.getTrackbarPos("SMax", "image")

            v_min = cv2.getTrackbarPos("VMin", "image")
            v_max = cv2.getTrackbarPos("VMax", "image")

            output = cv2.bitwise_and(
                img,
                img,
                mask=hsv_min_max_clip_mask(
                    img,
                    numpy.array([h_min, s_min, v_min]),
                    numpy.array([h_max, s_max, v_max]),
                ),
            )

            # Print if there is a change in HSV value
            if (
                (ph_min != h_min)
                | (ps_min != s_min)
                | (pv_min != v_min)
                | (ph_max != h_max)
                | (ps_max != s_max)
                | (pv_max != v_max)
            ):
                if verbose:
                    print(
                        f"(hMin = {h_min:d} , sMin = {s_min:d}, vMin = {v_min:d}), (hMax = {h_max:d} , sMax = {s_max:d}, vMax = {v_max:d})"
                    )

                # TODO :look at neater handling of min-maxing values
                th_min = min(h_min, h_max)
                ts_min = min(s_min, s_max)
                tv_min = min(v_min, v_max)
                h_max = max(h_min, h_max)
                s_max = max(s_min, s_max)
                v_max = max(v_min, v_max)
                h_min = th_min
                s_min = ts_min
                v_min = tv_min

                cv2.setTrackbarPos("HMin", "image", h_min)
                cv2.setTrackbarPos("HMax", "image", h_max)
                cv2.setTrackbarPos("SMin", "image", s_min)
                cv2.setTrackbarPos("SMax", "image", s_max)
                cv2.setTrackbarPos("VMin", "image", v_min)
                cv2.setTrackbarPos("VMax", "image", v_max)

                ph_min = h_min
                ps_min = s_min
                pv_min = v_min
                ph_max = h_max
                ps_max = s_max
                pv_max = v_max

            if show_image(output, "image", wait=waitTime):
                break

    cv2.destroyAllWindows()


if __name__ == "__main__":
    pss = (
        Path.home()
        / "ProjectsWin/AiBitbucket/Internal/OptikosPrime/exclude/new_images/adam_plus0p5/215asd.jpg",
        Path(r"C:\Users\Christian\OneDrive\Billeder\buh\7BIsT.png"),
        Path(r"C:\Users\Christian\OneDrive\Billeder\Portraits\thomas.jpg"),
    )
    if any(p.exists() for p in pss):
        interative_hsv_color_picker(iter(pss))
    else:
        print(f"{pss} does not exist")



import cv2
import numpy
from typing import Tuple, Union, Sequence

__all__ = ["circle_crop"]


def circle_crop(image: numpy.ndarray, center: Sequence, radius: int) -> numpy.ndarray:
    mask = numpy.zeros(image.shape[:2], dtype="uint8")
    cv2.circle(mask, center, radius, 255, -1)
    masked = cv2.bitwise_and(image, image, mask=mask)
    cropped = masked[
        center[1] - radius : center[1] + radius + 1,
        center[0] - radius : center[0] + radius + 1,
    ]

    return cropped



from typing import Optional

import cv2

from warg import sink


__all__ = ["add_trackbar"]


def add_trackbar(
    window_name: str,
    trackbar_name: str,
    default: Optional[int] = None,
    *,
    min_val: int = 0,
    max_val: int = 255,
    callback: callable = sink
) -> None:
    """
    Adds a trackbar to a window.
    :param window_name: Name of the window to add the trackbar to.
    :param trackbar_name: Name of the trackbar.
    :param trackbar_max: Maximum value of the trackbar.
    :param callback: Callback function to be called when the trackbar is moved.
    :return: None
    """
    if default is None:
        default = min_val + (max_val - min_val) // 2
    default = int(default)
    min_val = int(min_val)
    max_val = int(max_val)
    cv2.createTrackbar(trackbar_name, window_name, default, max_val, callback)
    # cv2.setTrackbarPos(trackbar_name, window_name, default)
    cv2.setTrackbarMin(trackbar_name, window_name, min_val)
    # cv2.setTrackbarMax(trackbar_name, window_name, max_val)
